from typing import Any, List

from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Manager, QuerySet
from django.utils.translation import gettext_lazy as _


class JSONObjectField(models.JSONField):  # type: ignore
    """
    A field for storing JSON Objects.

    Other JSON types are not allowed at the top level.
    """

    @staticmethod
    def _validate_is_object(val) -> None:
        """Validate that the value is a JSON Object (dict) at the top level."""
        if not isinstance(val, dict):
            raise ValidationError(_('Must be a JSON Object.'))

    empty_values: List[Any] = [{}]
    default_validators = [_validate_is_object]

    def __init__(self, *args, **kwargs):
        kwargs['default'] = dict
        kwargs['blank'] = True
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs['default']
        del kwargs['blank']
        return name, path, args, kwargs


class SelectRelatedManager(Manager):
    """
    A Manager which always follows specified foreign-key relationships.

    For Models which very frequently need to follow foreign-key relationships
    (e.g. to generate their own __str__), it may be best to make this the
    "_default_manager", so many automatically generated queries are more
    efficient. To do so, define a custom "objects" as the first
    Manager in the Model:
        objects = SelectRelatedManager('foreign_thing')

    Passing no arguments makes this follow all the Model's non-null
    foreign-key relationships.
    """

    def __init__(self, *related_fields: str) -> None:
        self.related_fields: List[str] = list(related_fields)
        super().__init__()

    def get_queryset(self) -> QuerySet:
        return super().get_queryset().select_related(*self.related_fields)


class DeferredFieldsManager(Manager):
    """A Manager which defers loading specified fields within fetched Models."""

    def __init__(self, *deferred_fields):
        self.deferred_fields = deferred_fields
        super().__init__()

    def get_queryset(self):
        return super().get_queryset().defer(*self.deferred_fields)



from datetime import timedelta

from django.core.files.storage import Storage

try:
    from storages.backends.s3boto3 import S3Boto3Storage
except ImportError:
    # This should only be used for type interrogation, never instantiation
    S3Boto3Storage = type('FakeS3Boto3Storage', (), {})
try:
    from minio_storage.storage import MinioStorage
except ImportError:
    # This should only be used for type interrogation, never instantiation
    MinioStorage = type('FakeMinioStorage', (), {})


def expiring_url(storage: Storage, name: str, expiration: timedelta) -> str:
    """
    Return an expiring URL to a file name on a `Storage`.

    `S3Boto3Storage` and `MinioStorage` are specifically supported.
    """
    # Each storage backend uses a slightly different API for URL expiration
    if isinstance(storage, S3Boto3Storage):
        return storage.url(name, expire=expiration.total_seconds())
    elif isinstance(storage, MinioStorage):
        return storage.url(name, max_age=expiration)
    else:
        # Unsupported Storage type
        return storage.url(name)



from django.contrib import admin


class ReadonlyTabularInline(admin.TabularInline):
    can_delete = False
    show_change_link = True
    view_on_site = False
    extra = 0

    def get_readonly_fields(self, request, obj=None):
        return self.fields

    def has_add_permission(self, request, obj=None):
        return False



from rest_framework.pagination import LimitOffsetPagination


class BoundedLimitOffsetPagination(LimitOffsetPagination):
    """
    Pagination class that forces pagination with an upper bound.

    DRF's built-in paginators such as LimitOffsetPagination will cause
    your endpoints to return different structure based on whether or not
    paging parameters were passed in the request. In most cases it's desirable
    to force pagination on endpoints that could return large result sets.
    Using this class has two benefits:

    1. It will always return responses in their paged structure, i.e. a
       JSON Object with "count", and a "results" Array. This is achieved by
       providing a default limit that is used if none is passed.
    2. It enforces a reasonable upper bound on page sizes. Without this, users could
       circumvent the intent of pagination by simply providing a high enough limit.

    The default limit will be the value of DRF's normal PAGE_SIZE setting, or
    the max limit if that is not set or falsy. The upper bound on the limit is 1000.
    """

    max_limit = 1000

    def get_limit(self, request) -> int:
        return super().get_limit(request) or self.max_limit



from contextlib import contextmanager
from pathlib import Path, PurePath
import shutil
import tempfile
from typing import Generator

from django.core.files import File
from django.db.models.fields.files import FieldFile


@contextmanager
def field_file_to_local_path(field_file: FieldFile) -> Generator[Path, None, None]:
    """
    Provide a filesystem path to the content of a FieldFile.

    Normally, a FieldFile is usable as a Python file-like object, as:
    `with model.field.open() as file_stream:`

    However, in some use cases, particularly when invoking third-party libraries,
    an actual filesystem path to the content is required. This context manager
    provides this as:
    `with field_file_to_local_path(model.field) as file_path:`

    When this context manager exists, the filesystem content and path will be
    garbage collected.
    """
    with field_file.open('rb'):
        file_obj: File = field_file.file

        if type(file_obj) is File:
            # When file_obj is an actual File, (typically backed by FileSystemStorage),
            # it is already at a stable path on disk.
            yield Path(file_obj.name)
        else:
            # When file_obj is actually a subclass of File, it only provides a Python
            # file-like object API. So, it must be copied to a stable path.
            field_file_basename = PurePath(field_file.name).name
            with tempfile.NamedTemporaryFile('wb', suffix=field_file_basename) as dest_stream:
                shutil.copyfileobj(file_obj, dest_stream)
                dest_stream.flush()

                yield Path(dest_stream.name)



from pathlib import Path

from setuptools import find_packages, setup

readme_file = Path(__file__).parent / 'README.md'
with readme_file.open() as f:
    long_description = f.read()

setup(
    name='django-girder-utils',
    description='Django utilities for data management applications',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='Apache 2.0',
    url='https://github.com/girder/django-girder-utils',
    project_urls={
        'Bug Reports': 'https://github.com/girder/django-girder-utils/issues',
        'Source': 'https://github.com/girder/django-girder-utils',
    },
    author='Kitware, Inc',
    author_email='kitware@kitware.com',
    keywords='django girder',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django :: 3.0',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python',
    ],
    python_requires='>=3.8',
    install_requires=['django>=3'],
    packages=find_packages(),
)
